import os
import asyncio
import functools
import logging
import time
import signal
import json

import paho.mqtt.client as mqtt

logger = logging.getLogger(__name__)


async def mqtt_send(future, channel_layer, channel, event):
    result = await channel_layer.send(channel, event)
    future.set_result(result)


async def mqtt_group_send(future, channel_layer, group, event):
    result = await channel_layer.group_send(group, event)
    future.set_result(result)


# Only for groups
async def mqtt_group_add(future, channel_layer, group):
    channel_layer.channel_name = channel_layer.channel_name or await channel_layer.new_channel()
    result = await channel_layer.group_add(group, channel_layer.channel_name)
    future.set_result(result)


# Only for groups
async def mqtt_group_discard(future, channel_layer, group):
    result = await channel_layer.group_discard(group, channel_layer.channel_name)
    future.set_result(result)


class Server(object):
    def __init__(self, channel, host, port, username=None, password=None,
                 client_id=None, topics_subscription=None, mqtt_channel_name=None,
                 mqtt_channel_sub=None, mqtt_channel_pub=None):

        self.channel = channel
        self.host = host
        self.port = port
        self.client_id = client_id
        self.client = mqtt.Client(client_id=self.client_id, userdata={
            "server": self,
            "channel": self.channel,
            "host": self.host,
            "port": self.port,
        })
        self.username = username
        self.password = password
        self.stop = False
        self.loop = None
        self.client.on_connect = self._on_connect
        self.client.on_disconnect = self._on_disconnect
        self.client.on_message = self._on_message

        self.topics_subscription = topics_subscription or [("#", 2), ]
        assert isinstance(self.topics_subscription, list), "Topic subscription must be a list with (topic, qos)"

        self.mqtt_channel_name = mqtt_channel_name or "mqtt"
        self.mqtt_channel_pub = mqtt_channel_pub or "mqtt.pub"
        self.mqtt_channel_sub = mqtt_channel_sub or "mqtt.sub"

    def _on_connect(self, client, userdata, flags, rc):
        logger.info("Connected with status {}".format(rc))
        client.subscribe(self.topics_subscription)

    def _on_disconnect(self, client, userdata, rc):
        logger.info("Disconnected")
        if not self.stop:
            j = 3
            for i in range(j):
                logger.info("Trying to reconnect")
                try:
                    client.reconnect()
                    logger.info("Reconnected")
                    break
                except Exception as e:
                    if i < j:
                        logger.warn(e)
                        time.sleep(1)
                        continue
                    else:
                        raise

    @staticmethod
    def _mqtt_send_got_result(future):
        logger.debug("Sending message to MQTT channel.")
        result = future.result()
        if result:
            logger.debug("Result: %s", result)

    def _on_message(self, client, userdata, message):
        logger.debug("Received message from topic {}".format(message.topic))
        payload = message.payload.decode("utf-8")

        try:
            payload = json.loads(payload)
        except Exception as e:
            logger.debug("Payload is not a JSON Serializable: {}".format(e.what()))
            pass

        logger.debug("Raw message {}".format(payload))

        # Compose a message for Channel with raw data received from MQTT
        msg = {
            "topic": message.topic,
            "payload": payload,
            "qos": message.qos,
            "host": userdata["host"],
            "port": userdata["port"],
        }

        try:
            # create a coroutine and send
            future = asyncio.Future()
            asyncio.ensure_future(
                mqtt_send(
                    future,
                    self.channel,
                    self.mqtt_channel_name,
                    {
                        "type": self.mqtt_channel_sub,
                        "text": msg
                    })
            )

            # attach callback for logs only
            future.add_done_callback(self._mqtt_send_got_result)

        except Exception as e:
            logger.error("Cannot send message {}".format(msg))
            logger.exception(e)

    async def client_pool_start(self):
        """
        This is the main loop pool for receiving MQTT messages
        """
        if self.username:
            self.client.username_pw_set(username=self.username, password=self.password)

        self.client.connect(self.host, self.port)

        logger.info("Starting loop")

        while True:
            self.client.loop(0.1)
            #  logger.debug("Restarting loop")
            await asyncio.sleep(0.1)

    def _mqtt_receive(self, msg):
        """
        Receive a message from Channel `mqtt.pub` and send it to MQTT broker
        """
        logger.info("Receive raw messages:\r\n%s", msg)
        # We only listen for messages from mqtt_channel_pub
        if msg['type'] == self.mqtt_channel_pub:

            payload = msg['text']

            if not isinstance(payload, dict):
                payload = json.loads(payload)

            logger.info("Receive a menssage with payload:\r\n%s", msg)
            self.client.publish(
                payload['topic'],
                payload['payload'],
                qos=payload.get('qos', 2),
                retain=False)

    async def client_pool_message(self):
        logger.info("Loop for messages pool")

        while True:
            logger.info("Wait for a message from channel %s", self.mqtt_channel_pub)
            self._mqtt_receive(await self.channel.receive(self.mqtt_channel_pub))

    def stop_server(self, signum):
        logger.info("Received signal {}, terminating".format(signum))
        self.stop = True
        for task in asyncio.Task.all_tasks():
            task.cancel()
        self.loop.stop()

    def run(self):
        self.stop = False
        loop = asyncio.get_event_loop()
        self.loop = loop

        for signame in ('SIGINT', 'SIGTERM'):
            loop.add_signal_handler(
                getattr(signal, signame),
                functools.partial(self.stop_server, signame)
            )

        print("Event loop running forever, press Ctrl+C to interrupt.")
        print("pid %s: send SIGINT or SIGTERM to exit." % os.getpid())

        tasks = asyncio.gather(*[
            asyncio.ensure_future(self.client_pool_start()),
            asyncio.ensure_future(self.client_pool_message()),
        ])

        asyncio.wait(tasks)

        try:
            loop.run_forever()
        finally:
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()

        self.client.disconnect()



import sys
import importlib
import argparse
import logging

from .server import Server


logger = logging.getLogger(__name__)


def channel_type(channel_name):
    sys.path.insert(0, ".")
    module_path, object_path = channel_name.split(":", 1)
    channel_layer = importlib.import_module(module_path)
    for bit in object_path.split("."):
        channel_layer = getattr(channel_layer, bit)

    return channel_layer


def main():
    parser = argparse.ArgumentParser(description="Interface between MQTT and ASGI and Channels 2.0 compatible")
    parser.add_argument("-H", "--host", help="MQTT broker host",
                        default="localhost")
    parser.add_argument("-p", "--port", help="MQTT broker port", type=int,
                        default=1883)
    parser.add_argument("-v", "--verbosity", action="count", default=0,
                        help="Set verbosity")
    parser.add_argument("channel_layer",
                        help=("The ASGI channel layer instance to use as "
                              "path.to.module:instance.path"))
    parser.add_argument("-U", "--username", help="MQTT username to authorised connection")
    parser.add_argument("-P", "--password", help="MQTT password to authorised connection")
    parser.add_argument("-i", "--id", dest="client_id", help="MQTT Cliente ID")

    parser.add_argument("--topic", action="append", dest="topics",
            help="MQTT topics with qos to subscribe \
                    --topic TOPIC:QOS \
                    --topic /office/sensor:0 --topic /home/sensor:1 \
                    If empty (#, 2) is set as default")

    parser.add_argument("-n", "--channel-name",
            help="Name of Channels's channel to send and receive messages")

    parser.add_argument("-s", "--channel-sub",
            help="Name of Channels's channel for MQTT Sub messages, default is mqtt.pub")

    parser.add_argument("-x", "--channel-pub",
            help="Name of Channels's channel for MQTT Pub messages, default is mqtt.sub")

    args = parser.parse_args()
 
    logging.basicConfig(
        level={
            0: logging.WARN,
            1: logging.INFO,
            2: logging.DEBUG,
        }.get(args.verbosity, logging.DEBUG),
        format="%(asctime)-15s %(levelname)-8s %(message)s",
    )

    channel_layer = channel_type(args.channel_layer)
    topics = []
    if args.topics:
        for t in args.topics:
            topic, qos = t.split(':')
            topics.append((topic, int(qos)))

    logger.info("\r\nStarting interface to MQTT broker {}:{}, channel {}\r\n\
            Topics: {}\r\n\
            MQTT channel name: {}\r\n\
            MQTT SUB: {}\r\n\
            MQTT PUB: {}\r\n".format(
                args.host, args.port, args.channel_layer,
                "\r\n".join(["{}, qos:{}".format(*t) for t in topics]),
                args.channel_name,
                args.channel_sub,
                args.channel_pub,
    ))

    server = Server(
            channel_layer, 
            args.host, 
            args.port, 
            args.username, 
            args.password,
            args.client_id,
            topics_subscription=topics, 
            mqtt_channel_name=args.channel_name, 
            mqtt_channel_sub=args.channel_sub, 
            mqtt_channel_pub=args.channel_pub
        )

    server.run()



import os.path
import re
from setuptools import setup, find_packages

def get_version(package):
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("^__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


setup(
    name="django_mqtt_bridge",
    version=get_version("django_mqtt_bridge"),
    author="Nilson Oliveira",
    author_email="nilsonmont.o@gmail.com",
    url="https://github.com/nilmonto/mqtt-django-bridge",
    description="Interface MQTT ASGI compatible with Django Channels 3.",
    long_description= open("README.md").read(),
    long_description_content_type="text/markdown",
    license="GPLv3+",
    packages=find_packages(),
    install_requires=[
        "paho-mqtt",
    ],
    entry_points={
        "console_scripts": [
            "django_mqtt_bridge=django_mqtt_bridge.cli:main",
        ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Framework :: Django',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
    ]
)
